//+------------------------------------------------------------------+
//|                                                     template.mq5 |
//|                                  Copyright 2022, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2022, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#property version   "1.00"
#define __ECON 4

#property            script_show_inputs

#include <ct_14_1s.mqh>

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct STableEvents
  {
      datetime                   date;
      string                     code;
      double                     actual;
      double                     forecast;
      double                     previous;
      int                        country_id;
      int                        currency_id;
      int                        event_id;
      
                                 STableEvents(){};
                                 ~STableEvents(){};
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct STableEventTypes
  {
      int                        id;
      string                     type;
      
                                 STableEventTypes(){};
                                 ~STableEventTypes(){};
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct STableCurrencies
  {
      int                        id;
      string                     currency;
      
                                 STableCurrencies(){};
                                 ~STableCurrencies(){};
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct STableCountries
  {
      int                        id;
      string                     country;
      
                                 STableCountries(){};
                                 ~STableCountries(){};
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CDatabase
  {
      public:
      
      STableEvents               events;
      STableEventTypes           event_types;
      STableCountries            countries;
      STableCurrencies           currncies;
      
      
                                 CDatabase(void);
                                 ~CDatabase(void);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum EType
  {
      TYPE_CPI=1,
      TYPE_PMI=2,
      TYPE_RETAIL_SALES=3,
      TYPE_AUCTION_10YR=4,
      TYPE_UNDEFINED=5,
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum ECountry
  {
      US=0,
      DE=1,
      GB=2,
      COUNTRY_UNDEFINED=3,
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
      CElement<string> _element_value;
      
      CMorphism<string,string> _morph_value;
      
      CCategory _category_economic,_category_sp500;
      
      CDomain<string> _domain_economic,_domain_sp500;
      
      CHomomorphism<string,string> _hmorph_economic,_hmorph_sp500;
      
      CElement<string> _e_retail,_e_cpi,_e_auction,_e_pmi;
//+------------------------------------------------------------------+
//| INPUTS                                                           |
//+------------------------------------------------------------------+
//country index 0 for US, 1 for Germany, and 2 for UK
input int __country=0;
input string __currency="USD";

//look back months
input int __bars=43;

// consant morph, the c value in equation y=mx + c, assuming a linear realtion between domain & codomain
input double __constant_morph=0.0;

input bool __functor_to_objects=true;
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
   {
      MqlRates _r[];
      
      ArraySetAsSeries(_r,true);
      
      double _forecast[];ArrayResize(_forecast,__bars-1);ArrayInitialize(_forecast,0.0);
      double _actual[];ArrayResize(_actual,__bars-1);ArrayInitialize(_actual,0.0);
      
      double _previous_actual=0.0,_previous_forecast=0.0;
            
      if(CopyRates(Symbol(),Period(),0,__bars,_r)>=__bars)
      {
         ArraySetAsSeries(_r,true);
         
         for(int i=__bars-1;i>0;i--)
         {
            _actual[i-1]=(_r[i-1].high-_r[i-1].low)/Point();
            
            if(__functor_to_objects)
            {
               double _f=GetObjectsForecast(_r[i]);
               if(_f<=0.0)
               {
                  _forecast[i-1]=NormalizeForecast(_f,(_r[i].high-_r[i].low)/Point(),_previous_actual,_previous_forecast);
               }
               else
               {
                  _forecast[i-1]=_f;
               }
               
               printf(__FUNCSIG__+" Objects forecast is: "+DoubleToString(_forecast[i-1])+", on: "+TimeToString(_r[i].time)+", with actual being: "+DoubleToString(_actual[i-1]));
            }
            else if(!__functor_to_objects)
            {
               double _f=GetMorphsForecast(_r[i]);
               if(_f<=0.0)
               {
                  _forecast[i-1]=NormalizeForecast(_f,(_r[i].high-_r[i].low)/Point(),_previous_actual,_previous_forecast);
               }
               else
               {
                  _forecast[i-1]=_f;
               }
               
               printf(__FUNCSIG__+" Morphs forecast is: "+DoubleToString(_forecast[i-1])+", on: "+TimeToString(_r[i].time)+", with actual being: "+DoubleToString(_actual[i-1]));
            }
            
            _previous_actual=(_r[i].high-_r[i].low)/Point();
            _previous_forecast=_forecast[i-1];
         }
         
         double _r=0.0;
         MathCorrelationSpearman(_forecast,_actual,_r);
         
         printf(__FUNCSIG__+" correlation is: "+DoubleToString(_r)+" with morph constant as: "+DoubleToString(__constant_morph));
      }
   } 
//+------------------------------------------------------------------+
//| Get Forecast value, forecast for next change in price bar range. |
//+------------------------------------------------------------------+
double NormalizeForecast(double NegativeValue,double CurrentActual,double PreviousActual,double PreviousForecast)
   {
      return(CurrentActual+(((CurrentActual-NegativeValue)/CurrentActual)*fabs(PreviousActual-PreviousForecast)));
   }
//+------------------------------------------------------------------+
//| Get Forecast value, forecast for next change in price bar range. |
//+------------------------------------------------------------------+
double GetObjectsForecast(MqlRates &R)
   {
      double _output=0.0;
      
      //econ init
      SetEconomic(R);
      
      //sp500 init
      SetSP500(R);
      
      matrix _domain;
      vector _codomain,_inputs;
      
      _domain.Init(__ECON+1,__ECON);
      
      for(int r=0;r<__ECON+1;r++)
      {
         CDomain<string> _d;_d.Let();
         _category_economic.GetDomain(_category_economic.Domains()-r-1,_d);
         
         for(int c=0;c<__ECON;c++)
         {
            CElement<string> _e; _d.Get(c,_e);
            
            string _s; _e.Get(1,_s);
            
            _domain[r][c]=StringToDouble(_s);
         }
      }
      
      _codomain.Init(__ECON);
      
      for(int r=0;r<__ECON;r++)
      {
         CDomain<string> _d;_d.Let();
         _category_sp500.GetDomain(_category_sp500.Domains()-r-1,_d);
         
         CElement<string> _e; _d.Get(0,_e);
         
         string _s; _e.Get(0,_s);
            
         _codomain[r]=StringToDouble(_s);
      }
      
      _inputs.Init(__ECON);_inputs.Fill(__constant_morph);
      
      M(_domain,_codomain,_inputs,_output,1);
      
      return(_output);
   }
//+------------------------------------------------------------------+
//|   Get Forecast value, forecast for next change in price bar range. |
//+------------------------------------------------------------------+
double GetMorphsForecast(MqlRates &R)
   {
      double _output=0.0;
      
      //econ init
      SetEconomic(R);
      
      //sp500 init
      SetSP500(R);
      
      matrix _domain;
      vector _codomain,_inputs;
      
      _domain.Init(__ECON+1,__ECON);
      
      for(int r=0;r<__ECON+1;r++)
      {
         CDomain<string> _d_new,_d_old;
         _category_economic.GetDomain(_category_economic.Domains()-r-1,_d_new);
         
         if(_category_economic.Domains()-r-1-1>=0){ _category_economic.GetDomain(_category_economic.Domains()-r-1-1,_d_old); }
         
         for(int c=0;c<__ECON;c++)
         {
            CElement<string> _e_new; _d_new.Get(c,_e_new);
            
            string _s_new; _e_new.Get(1,_s_new);
            
            CElement<string> _e_old; _d_old.Get(c,_e_old);
            
            string _s_old; _e_old.Get(1,_s_old);
            
            _domain[r][c]=StringToDouble(_s_new)-StringToDouble(_s_old);
         }
      }
      
      _codomain.Init(__ECON);
      
      for(int r=0;r<__ECON;r++)
      {
         CDomain<string> _d_new,_d_old;
         _category_sp500.GetDomain(_category_sp500.Domains()-r-1,_d_new);
         
         if(_category_sp500.Domains()-r-1-1>=0){ _category_sp500.GetDomain(_category_sp500.Domains()-r-1-1,_d_old); }
         
         CElement<string> _e_new; _d_new.Get(0,_e_new);
         
         string _s_new; _e_new.Get(0,_s_new);
         
         CElement<string> _e_old; _d_old.Get(0,_e_old);
         
         string _s_old; _e_old.Get(0,_s_old);
         
         _codomain[r]=StringToDouble(_s_new)-StringToDouble(_s_old);
      }
      
      _inputs.Init(__ECON);_inputs.Fill(__constant_morph);
      
      M(_domain,_codomain,_inputs,_output,1);
      
      return(_output);
   }
//+------------------------------------------------------------------+
//|   Get Ocean data from file defined in input.                     | 
//|   Load it into Ocean category.                                   |
//+------------------------------------------------------------------+
void SetEconomic(MqlRates &R)
   {
      datetime _time=R.time;
      
      int _elements=1;
      
      _e_retail.Let();
      SampleEvents(_time,_e_retail,IntToCountry(__country),__currency,TYPE_RETAIL_SALES);//printf(__FUNCSIG__+"...retail. ");
      
      _e_cpi.Let();
      string _cpi_time="";_e_retail.Get(0,_cpi_time);
      SampleEvents(StringToTime(_cpi_time),_e_cpi,IntToCountry(__country),__currency,TYPE_CPI);//printf(__FUNCSIG__+"...cpi. ");
      
      _e_auction.Let();
      string _auction_time="";_e_cpi.Get(0,_auction_time);
      SampleEvents(StringToTime(_auction_time),_e_auction,IntToCountry(__country),__currency,TYPE_AUCTION_10YR);//printf(__FUNCSIG__+"...auction. ");
      
      _e_pmi.Let();
      string _pmi_time="";_e_auction.Get(0,_pmi_time);
      SampleEvents(StringToTime(_pmi_time),_e_pmi,IntToCountry(__country),__currency,TYPE_PMI);//printf(__FUNCSIG__+"...pmi. ");
      
       _domain_economic.Cardinality(__ECON);
       _domain_economic.Set(0,_e_retail);
       _domain_economic.Set(1,_e_cpi);
       _domain_economic.Set(2,_e_auction);
       _domain_economic.Set(3,_e_pmi);
      
       _category_economic.SetDomain(_category_economic.Domains(),_domain_economic);
   }
//+------------------------------------------------------------------+
//|   Get SP500-100 data from symbol (NDX100 for this broker).       | 
//|   Load it into SP500 category.                                   |
//+------------------------------------------------------------------+
void SetSP500(MqlRates &R)
   {
      _hmorph_sp500.Let();
      
      double _float_value=0.0;
      
      _float_value=(R.high-R.low)/Point();
      
      _element_value.Let();_element_value.Cardinality(1);_element_value.Set(0,DoubleToString(_float_value));
      _domain_sp500.Cardinality(1);_domain_sp500.Set(0,_element_value);
      
      //_category_sp500.Domains(_category_sp500.Domains()+1);
      _category_sp500.SetDomain(_category_sp500.Domains(),_domain_sp500);  
   }
//+------------------------------------------------------------------+
//| Method to retieve Calendar Events by Type and by Currency.       |
//| Events parameter is an output counter to no. of added events to  |
//| domain.                                                          |
//+------------------------------------------------------------------+
void SampleEvents(datetime Start,CElement<string> &E,int CountryID,string Currency,EType Type)
   {
      //datetime _start_time=Start;
      datetime _start_time=datetime(int(Start)-int(12* PeriodSeconds(PERIOD_MN1)));
      //
      
      E.Let();
      E.Cardinality(2);
      
//--- get values
      MqlCalendarValue _value[];
      datetime _stop_date=Start;
//--- get events
      ResetLastError();
      MqlCalendarEvent _event[];
      int _new_events=CalendarEventByCurrency(Currency,_event);
      
      bool _found=false;
      //
      for(int e=0;e<_new_events;e++)
      {
         int _values=CalendarValueHistoryByEvent(_event[e].id, _value, _start_time, _stop_date);
         //
         if(_event[e].country_id!=CountryID){ continue; }
         //
         for(int v=0;v<_values;v++)
         {
            //data population at each index
            
            if(_value[v].HasActualValue())
            {
               int _index=TypeToInt(_event[e].event_code);
               
               string _type=IntToType(_index);
               
               if(_type==EnumToString(Type))
               {
                  E.Set(1,DoubleToString(_value[v].GetActualValue()));
                  
                  E.Set(0,TimeToString(_value[v].time));
         
                  //printf(__FUNCSIG__+" code: "+_event[e].event_code+", has value: "+DoubleToString(_value[v].GetActualValue()));
                  
                  _found=true; break;
               }
            }
            
            if(_found){ break; }
         }
         
         if(_found){ break; }
      }
   }
//+------------------------------------------------------------------+
//|   Model on Event Type Classification                             |
//+------------------------------------------------------------------+
int EventType(string EventCode)
   {
      if(StringFind(EventCode,"cpi")>=0){ return(0); }
      else if(StringFind(EventCode,"pmi")>=0){ return(1); }
      else if((StringFind(EventCode,"2-yr")>=0||StringFind(EventCode,"2-year")>=0)&&StringFind(EventCode,"auction")>=0){ return(2); }
      else if((StringFind(EventCode,"10-yr")>=0||StringFind(EventCode,"10-year")>=0)&&StringFind(EventCode,"auction")>=0){ return(3); }
      else if(StringFind(EventCode,"interest")>=0){ return(4); }
      //... other options can be added here, based on training.
      
      return(5);
   }
//+------------------------------------------------------------------+
//| Method that converts a currency to its index in the Currency set |
//+------------------------------------------------------------------+
int EventCurrency(string Currency)
   {
      if(Currency=="EUR"){ return(0); }
      else if(Currency=="GBP"){ return(1); }
      else if(Currency=="USD"){ return(2); }
      else if(Currency=="CHF"){ return(3); }
      else if(Currency=="JPY"){ return(4); }
      
      return(-1);
   }
//+------------------------------------------------------------------+
//|   Event Type To Int Classifier                                   |
//+------------------------------------------------------------------+
int TypeToInt(string EventCode)
   {
      if(StringFind(EventCode,"cpi")>=0 && StringFind(EventCode,"mm")>=0){ return(0); }
      else if(StringFind(EventCode,"pmi")>=0 && StringFind(EventCode,"manufacturing")>=0){ return(1); }
      else if(StringFind(EventCode,"retail")>=0 && StringFind(EventCode,"sales")>=0 && StringFind(EventCode,"mm")>=0){ return(2); }
      else if(StringFind(EventCode,"10-")>=0 && StringFind(EventCode,"auction")>=0){ return(3); }
      
      return(4);
   }
//+------------------------------------------------------------------+
//|   Event Int To Type Classifier                                   |
//+------------------------------------------------------------------+
string IntToType(int EventIndex)
   {
      if(EventIndex==0){ return(EnumToString(TYPE_CPI)); }
      else if(EventIndex==1){ return(EnumToString(TYPE_PMI)); }
      else if(EventIndex==2){ return(EnumToString(TYPE_RETAIL_SALES)); }
      else if(EventIndex==3){ return(EnumToString(TYPE_AUCTION_10YR)); }
      
      return(EnumToString(TYPE_UNDEFINED));
   }
//+------------------------------------------------------------------+
//|   Event Int To Country Classifier                                |
//+------------------------------------------------------------------+
int IntToCountry(int EventIndex)
   {
      if(EventIndex==0){ return(840); }
      else if(EventIndex==1){ return(276); }
      else if(EventIndex==2){ return(826); }
      
      return(-1);
   }
//+------------------------------------------------------------------+
